@@ -12,9 +12,12 @@ use core::hash::{Hash, Hasher};
 use core::mem;
 use core::num::NonZeroU16;
 use core::ops::Deref;
-use hashbrown::HashMap;
+use hashbrown::{HashMap, HashSet};
 use ttf_parser::{Face, FaceParsingError, GlyphId, Tag};
 
+#[cfg(feature = "parallel")]
+use rayon::prelude::*;
+
 /// Defines the bounds for a glyph's outline in subpixels. A glyph's outline is always contained in
 /// its bitmap.
 #[derive(Copy, Clone, PartialEq, Debug)]
@@ -131,6 +134,7 @@ impl LineMetrics {
     }
 }
 
+/// Stores compiled geometry and metric information.
 #[derive(Clone)]
 pub(crate) struct Glyph {
     pub v_lines: Vec<Line>,
@@ -200,6 +204,7 @@ impl core::fmt::Debug for Font {
             .field("name", &self.name)
             .field("settings", &self.settings)
             .field("units_per_em", &self.units_per_em)
+            .field("hash", &self.hash)
             .finish()
     }
 }
@@ -245,13 +250,17 @@ impl Font {
         })();
 
         // Collect all the unique codepoint to glyph mappings.
-        let mut char_to_glyph = HashMap::new();
+        let glyph_count = face.number_of_glyphs();
+        let mut seen_mappings = HashSet::with_capacity(glyph_count as usize);
+        let mut char_to_glyph = HashMap::with_capacity(glyph_count as usize);
+        seen_mappings.insert(0u16);
         for subtable in face.character_mapping_subtables() {
             subtable.codepoints(|codepoint| {
                 if let Some(mapping) = subtable.glyph_index(codepoint) {
-                    // Zero is a valid value for missing glyphs, so even if a mapping is zero, the
-                    // result is desireable.
-                    char_to_glyph.insert(codepoint, unsafe { NonZeroU16::new_unchecked(mapping.0) });
+                    if let Some(mapping) = NonZeroU16::new(mapping.0) {
+                        seen_mappings.insert(mapping.get());
+                        char_to_glyph.insert(codepoint, mapping);
+                    }
                 }
             });
         }
@@ -261,16 +270,15 @@ impl Font {
         let units_per_em = face.units_per_em().unwrap_or(1000) as f32;
 
         // Parse and store all unique codepoints.
-        let glyph_count = face.number_of_glyphs() as usize;
-        let mut glyphs: Vec<Glyph> = vec::from_elem(Glyph::default(), glyph_count);
-        for (_, mapping) in &char_to_glyph {
-            let mapping = unsafe { mem::transmute::<NonZeroU16, u16>(*mapping) as usize };
-            if mapping as usize >= glyph_count {
+        let mut glyphs: Vec<Glyph> = vec::from_elem(Glyph::default(), glyph_count as usize);
+
+        let generate_glyph = |index: u16| -> Result<Glyph, &'static str> {
+            if index >= glyph_count {
                 return Err("Attempted to map a codepoint out of bounds.");
             }
 
-            let glyph_id = GlyphId(mapping as u16);
-            let glyph = &mut glyphs[mapping];
+            let mut glyph = Glyph::default();
+            let glyph_id = GlyphId(index);
             if let Some(advance_width) = face.glyph_hor_advance(glyph_id) {
                 glyph.advance_width = advance_width as f32;
             }
@@ -280,7 +288,24 @@ impl Font {
 
             let mut geometry = Geometry::new(settings.scale, units_per_em);
             face.outline_glyph(glyph_id, &mut geometry);
-            geometry.finalize(glyph);
+            geometry.finalize(&mut glyph);
+            Ok(glyph)
+        };
+
+        #[cfg(not(feature = "parallel"))]
+        for index in seen_mappings {
+            glyphs[index as usize] = generate_glyph(index)?;
+        }
+
+        #[cfg(feature = "parallel")]
+        {
+            let generated: Vec<(u16, Glyph)> = seen_mappings
+                .into_par_iter()
+                .map(|index| Ok((index, generate_glyph(index)?)))
+                .collect::<Result<_, _>>()?;
+            for (index, glyph) in generated {
+                glyphs[index as usize] = glyph;
+            }
         }
 
         // New line metrics.
